% Social Golfer problem
% http://4c.ucc.ie/~tw/csplib//prob/prob010/index.html
% Copyright (C) 2003 INRIA Sylvain Soliman

% file soc_golf2.pl
% January 23 2003: Version 0.0.2

golf(G,S,W,R) :-     % G groups of S golfers play for W weeks
	M is G*S,
	length(R,W),      % R, the result, is a list of weeks, each being a list...
	%  R = [Week1|WL],
	%  instantiate_week1(Week1,G,S),
	init_weeks(M,G,S,R,Vars), % ... of group numbers attributed to each player
	not_twice_rec(R),
	randomize,
	fd_labeling(Vars,[value_method(random)]).%,[value_method(random)]). 
   
instantiate_week1([],0,_).
instantiate_week1(W,G,S) :-
	G>0,
	GG is G-1,
	length(L,S),
	fd_domain(L,G,G),
	instantiate_week1(LL,GG,S),
	append(LL,L,W).
	
golf_sbds(G,S,W,R) :-    
	fd_set_vector_max(16),
	M is G*S,
	length(R,W),  
	init_weeks(M,G,S,R,Vars),
	not_twice_rec(R),
	
	% Symmetry breaking
	LSym=[],
%	length(Vars,NbVars),
%	liste_num(IVars,1,NbVars),
%	liste_num(IGroup,1,G),
%	make_directory('./tmp/'),
%	open('./tmp/sym_val.pl',write,SymStream),
%	generate_sym_within_W2(IVars,IGroup,LSymWG,M,SymStream),
%	close(SymStream),
%	consult('./tmp/sym_val.pl'),
%	g_assign(cpt_permutation_week,0),
%	generate_sym_within_S(IVars,M,LSymWS),
%	append(LSymWG,LSymWS,LSym2),
%	generate_sym_player(IVars,M,LSymPlayer),
%	append(LSymPlayer,LSym2,LSym),
%	delete_file('./tmp/sym_val.pl'),
%	delete_directory('./tmp/'), 
	labeling_sbds(Vars,LSym,[],[val_order(min)]).

% Dynamic symmetries breaking of groups within each week
generate_sym_within_W([],_,[],_).
generate_sym_within_W(IVars,IGroup,LSymWG,M):-
	length(IVarsW,M),
	append(IVarsW,LIVars,IVars),
	SymWG=[sym_val(IVarsW,IGroup)],%sym_var(IVarsW,IGroup)],
	generate_sym_within_W(LIVars,IGroup,LSymWG2,M),
	append(SymWG,LSymWG2,LSymWG).
%%
generate_sym_within_W2([],_,[],_,_):-!.
generate_sym_within_W2(IVars,IGroup,LSymWG,M,SymStream):-
	length(IVarsW,M),
	append(IVarsW,LIVars,IVars),
	nth(1,IVarsW,Min),last(IVarsW,Max),
	predicats_sym(Min,Max,IGroup,LSymWG1,SymStream),
	generate_sym_within_W2(LIVars,IGroup,LSymWG2,M,SymStream),
	append(LSymWG1,LSymWG2,LSymWG).
	
predicats_sym(_,_,[_],[],_):-!.
predicats_sym(Min,Max,[Val1|IGroup],LSymWG,SymStream):-
	write_predicats_sym(Min,Max,Val1,IGroup,LSymWG1,SymStream),
	predicats_sym(Min,Max,IGroup,LSymWG2,SymStream),
	append(LSymWG1,LSymWG2,LSymWG).

write_predicats_sym(_,_,_,[],[],_).
write_predicats_sym(Min,Max,Val1,[Val2|IGroup],[Sym|LSymWG],SymStream):-
	number_atom(Val1,Suffixe1),
	number_atom(Val2,Suffixe2),
	number_atom(Min,Suffixe3),
	atom_concat('sym_var_',Suffixe1,Prefixe1),
	atom_concat(Prefixe1,'_',Prefixe2),
	atom_concat(Prefixe2,Suffixe2,Prefixe3),
	atom_concat(Prefixe3,'_',Prefixe4),
	atom_concat(Prefixe4,Suffixe3,Sym),
	
	format(SymStream,
		'~a(_,_,IVar,Val1,IVar,Val2):-\n\tIVar>~d,\n\tIVar<~d,\n\tSom is ~d+~d,\n\tVal2 is Som-Val1.\n\n',
		[Sym,Min,Max,Val1,Val2]),
	write_predicats_sym(Min,Max,Val1,IGroup,LSymWG,SymStream).
	

% Dynamic symmetries breaking of weeks within each solution
generate_sym_within_S([],_,[]).
generate_sym_within_S(IVars,M,LSymWS):-
	length(IVarsW,M),
	append(IVarsW,LIVars,IVars),
	week_permutation(IVarsW,LIVars,M,LSymWS1),
	append(LSymWS1,LSymWS2,LSymWS),
	generate_sym_within_S(LIVars,M,LSymWS2).

week_permutation(_,[],_,[]).
week_permutation(IVars,LIVars,M,[PermutationName|LSymWS2]):-
	length(IVars2,M),
	append(IVars2,LIVars2,LIVars),
	generate_week_permutation(IVars,IVars2,LPermutations),
	g_read(cpt_permutation_week,Cpt),
	Cpt2 is Cpt+1,
	g_assign(cpt_permutation_week,Cpt2),
	number_atom(Cpt,Suffixe),
	atom_concat('perm_week_',Suffixe,PermutationName),
	permutation_var(LPermutations,PermutationName,[],_),
	week_permutation(IVars,LIVars2,M,LSymWS2).
	
generate_week_permutation([],[],[]).
generate_week_permutation([IVar1|LIVar1],[IVar2|LIVar2],[IVar1-IVar2|LPermutation]):-
	generate_week_permutation(LIVar1,LIVar2,LPermutation).

% Dynamic symmetries breaking of weeks within each solution
generate_sym_player([],_,_).
generate_sym_player(LIVars,M,LSymPlayer):-
	length(IVars,M),
	append(IVars,LIVars2,LIVars),
	g_assign(sym_player,0),	
	player_permutation(IVars,LSymPlayer),
	generate_sym_player(LIVars2,M,LSymPlayer).

player_permutation([_],[]).
player_permutation([Var|LIVars],LSymPlayer):-
	all_permutation_player(Var,LIVars,LSymPlayer1),
	player_permutation(LIVars,LSymPlayer2),
	append(LSymPlayer1,LSymPlayer2,LSymPlayer).

all_permutation_player(_,[],[]).
all_permutation_player(Var1,[Var2|LIVars],[PermutationName|LSymPlayer]):-
	(var(PermutationName)->
		g_read(sym_player,NbSym),
		NbSym2 is NbSym+1,
		g_assign(sym_player,NbSym2),
		number_atom(NbSym,Suffixe),
		atom_concat('sym_player_',Suffixe,PermutationName)
		;
		true),
	permutation_var([Var1-Var2],PermutationName,[],_),
	all_permutation_player(Var1,LIVars,LSymPlayer).


	
% Initialization
init_weeks(_,_,_,[],[]).
init_weeks(M,G,S,[GL|WL],V) :-
	length(GL,M),           % M players in the list
	fd_domain(GL,1,G),      % each player gets a group number
	fd_exactly_rec(GL,G,S), % exactly S golfers in each group
	init_weeks(M,G,S,WL,V1),
	append(GL,V1,V).

fd_exactly_rec(_,0,_).
fd_exactly_rec(L,G,S) :-
   G > 0,
   GG is G-1,
   fd_exactly(S,L,G),      % exactly S golfers in group G
   fd_exactly_rec(L,GG,S).

not_twice_rec([]).
not_twice_rec([W|WL]) :-
   not_twice(W,WL),
   not_twice_rec(WL).

not_twice(_,[]).
not_twice(W,[H|T]) :-
   no_common_rec(W,H),
   not_twice(W,T).

no_common_rec([],[]).
no_common_rec([H1|T1],[H2|T2]) :-
   no_common(H1,T1,H2,T2),
   no_common_rec(T1,T2).

no_common(_,[],_,[]).
no_common(G1,[H1|T1],G2,[H2|T2]) :-
   (H1 #= G1) #==> (G2 #\= H2),
   (H2 #= G2) #==> (G1 #\= H1),
   no_common(G1,T1,G2,T2).
   
% Result:

% golf(5,3,7,R):
% portnb(newVariable, 8925).
% portnb(newConstraint, 13370).
% nbcstr(31045).

% golf_sbds(5,3,7,R):
% portnb(newVariable, 8925).
% portnb(newConstraint, 13370).
% nbcstr(31045).

